from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from functools import partial

from updaters import *
from imageProcessing import *
from taskman import *

import numpy as np
import time
import sys
import queue as queue

# TODO: get a threadpool on the go so you dont need to artifically keep the thread alive
# TODO: use pythons 'queue' to move settings into the task managet in a timely fashion
# TODO: maybe use a mutex to share the array with the canvas? maybe not necc.
##==============EngineOperator===============##
##===========================================##
# This is the interface between the GUI and the threads.
# Controls the Updater thread and the CanvasThread
class EngineOperator(QObject):
    settingsSig = pyqtSignal(dict)
    interruptSig = pyqtSignal()
    backgroundSig = pyqtSignal()
    gifresetSig = pyqtSignal()
    plainSig = pyqtSignal(int, int)

    def __init__(self, canvas, frameLabel, arrayfpsLabel, canvasfpsLabel, **kwargs):
    # The kwargs consists of the following: speed, updates, frames, beta,
    # stochastic?, threshold/coverage.
        QObject.__init__(self)
        self.kwargs = kwargs
        self.canvas = canvas
        self.frameLabel = frameLabel
        self.arrayfpsLabel = arrayfpsLabel
        self.canvasfpsLabel = canvasfpsLabel
        self.mainUpdates = 0

        self.taskman_init()

#=============Thread Communicators======#
#   def update_kwargs(self, **kwargs):
#       print('Updating Threads')
#       for i in kwargs:
#           self.kwargs[i] = kwargs[i]
#       if self.taskthread.isRunning():
#           print('Requesting Interrution')
#           self.taskthread.requestInterruption()
#       else:
#           self.settingsSig.emit({i:self.kwargs[i] for i in self.kwargs})

    def update_kwargs(self, **kwargs):
        for i in kwargs:
            self.kwargs[i] = kwargs[i]
        self.settingsSig.emit({i:self.kwargs[i] for i in self.kwargs})

    def reset_gifcount(self):
        self.gifresetSig.emit()

    def breaker(self):
        print('Breaker!')
        self.update_kwargs(RUN=False)
        self.update_kwargs(CLEAR=False)
        self.update_kwargs(EQUILIRATE=False)

    def thread_looper(self):
#       if self.kwargs['RUN'] is True and self.kwargs['RUNFRAMES'] < self.kwargs['IMAGEUPDATES']:
#           self.settingsSig.emit({i:self.kwargs[i] for i in self.kwargs})
#           self.taskthread.start()
#       elif self.kwargs['CLEAR'] is True or self.kwargs['EQUILIBRATE'] is True:
#           self.settingsSig.emit({i:self.kwargs[i] for i in self.kwargs})
#           self.taskthread.start()
#           self.simple_update(CLEAR=False, EQUILIBRATE=False)
#       else:
#           self.simple_update(RUN=False)
        print('Thread standing by.')

#===============Run Initiators=============#
    def static_run(self):
        self.taskthread.requestInterruption()
        self.update_kwargs(RUN=True, EQUILIBRATE=False, CLEAR=False, RUNFRAMES=(self.kwargs['IMAGEUPDATES']-1))
        self.taskthread.start()

    def dynamic_run(self):
        self.taskthread.requestInterruption()
        self.update_kwargs(RUN=True, EQUILIBRATE=False, CLEAR=False, RUNFRAMES=0)
        self.taskthread.start()

    def long_run(self):
        self.taskthread.requestInterruption()
        self.update_kwargs(RUN=False, EQUILIBRATE=True, CLEAR=False)
        self.taskthread.start()

    def clear_background(self):
        self.settingsSig.emit({i:self.kwargs[i] for i in self.kwargs})
        self.backgroundSig.emit()
        self.imagethread.start()

    def clear_array(self):
        self.taskthread.requestInterruption()
        self.update_kwargs(RUN=False, CLEAR=True, EQUILIBRATE=False, RUNFRAMES=(self.kwargs['IMAGEUPDATES']-2))
        self.plainSig.emit(self.kwargs['N'], self.kwargs['D'])
        self.imagethread.start()
        self.taskthread.start()

    def noise_array(self):
        pass
        self.taskthread.start()

#===============GUI updaters=============#
    def array_fps_update(self, value):
        if value:
            self.arrayfpsLabel.setText('Array fps: {:03.3f}'.format(1 / value))

    def canvas_fps_update(self, value):
        if value:
            self.canvasfpsLabel.setText('Canvas fps: {:03.3f}'.format(1 / value))

    def frame_value_update(self, value):
        self.frameLabel.setText(str(value))
        if not value % 10:
            self.kwargs['RUNFRAMES'] = value

    def error_string(self, error='Unlabelled Error! Oh no!'):
        print(error)

#=================Thread Initiator============#
#==================HERE BE DRAGONS============#
    def taskman_init(self):
        # Initialise the threads and the workers, and put them in place
        self.taskthread = QThread()
        self.imagethread = QThread()
        self.updatethread = QThread()
        self.handler = Handler(**self.kwargs)
        self.taskman = RunController(**self.kwargs)
        self.image = ImageCreator(**self.kwargs)
        self.taskman.moveToThread(self.taskthread)
        self.handler.moveToThread(self.updatethread)
        self.image.moveToThread(self.imagethread)

        self.taskthread.started.connect(self.taskman.process)
        self.taskthread.started.connect(self.imagethread.start)
        self.taskthread.started.connect(self.updatethread.start)

        # Connections from the engine to the workers
        self.settingsSig.connect(self.taskman.change_settings)
        self.settingsSig.connect(self.image.change_settings)
        self.interruptSig.connect(self.breaker)
        self.backgroundSig.connect(self.image.wolfram_paint)
        self.plainSig.connect(self.image.resize_array)
        self.gifresetSig.connect(self.image.reset_gifcount)

        # Connect up the signals between the workers
        self.taskman.handlerSig.connect(self.handler.next_array)
#       self.taskman.handlerinitSig.connect(self.handler.updater_start)
        self.handler.arraySig.connect(self.image.process)
#       self.handler.arrayinitSig.connect(self.image.processer_start)
        self.image.nextarraySig.connect(self.taskman.next_frame)
#       self.handler.nextSig.connect(self.taskman.next_frame)

        # Connections for closing threads WORK NECC HERE
        # Need to figure out exactly ho long a thread will stay waiting, what activates
        # it, how it proecesses signals it has recieved while it has been shutdown etc.
        # TODO
        self.taskman.finished.connect(self.taskthread.quit)
        self.taskman.finished.connect(self.updatethread.quit)
        self.taskman.finished.connect(self.imagethread.quit)
        self.taskthread.finished.connect(self.thread_looper)
        self.image.breakSig.connect(self.breaker)
        self.taskman.breakSig.connect(self.breaker)

        # Signals from the workers back to the GUI
        self.taskman.frameSig.connect(self.frame_value_update)
        self.handler.arrayfpsSig.connect(self.array_fps_update)
        self.image.canvasfpsSig.connect(self.canvas_fps_update)
        self.taskman.error.connect(self.error_string)
        self.image.error.connect(self.error_string)
        self.handler.error.connect(self.error_string)
        self.image.imageSig.connect(self.canvas.paint)

#       self.taskthread.start()



#!/usr/bin/python3

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

from GUI import MainWindow
import random as ra


def initVars():
    colHex1 = int(ra.random() * int('0xffffffff', 16))
    colHex2 = int(ra.random() * int('0xffffffff', 16))
    colHex3 = int(ra.random() * int('0xffffffff', 16))
    colHex4 = int(ra.random() * int('0xffffffff', 16))
    colHex5 = int('0xffffffff', 16)
    colHex6 = int('0xffffffff', 16)
    DEF = {
        'COLORLIST':    [colHex1, colHex2, colHex3, colHex4, colHex5, colHex6],
        'SATURATION':   80,     # This leaves changes on the image shortly
        'N':            300,    # Array dimensions
        'D':            300,    # Array dimensions
        'SCALE':        2,      # Image dim = N*SCALE x N*SCALE
        'BETA':         1 / 8,  # Critical temp for Ising
        'SPEED':        100,    # Throttle %
        'DEGREE':       4,      # Degree of the Potts model
        'IMAGEUPDATES': -1,    # Max number of frames to run
        'RUNFRAMES':    0,      # Frames in current run
        'MONTEUPDATES': 333,    # MonteCarlo updates per frame
        'LONGNUM':      100000, # MonteCarlo update to equilibrium
        'THRESHOLD':     0.1,   # Threshold for clear/noise (sigmoid function)
        'NEWARR':       1,      # New array upon engine creation?
        'STOCHASTIC':   True,   # Noise on?
        'CONWAY':       True,   # Conway on?
        'EQUILIBRATE':  False,  # Equilibrate array?
        'RUN':          False,   # Run the simulation
        'CLEAR':        False,  # Clear array?
                                # Update rules for conway
        'RULES':        [[3,6,2,2],\
                         [3,6,3,3],\
                         [3,4,3,3],\
                         [3,4,2,2]],
        'INTERRUPT':    False,  # Used to interrupt a run
        'WOLFRULE':     30,
        'WOLFWAVE':     True,
        'WOLFSCALE':    2,
        'WOLFPOLARITY': 1,
        'UB':           -1,
        'RB':           -1,
        'DB':           -1,
        'LB':           -1,
        'RECORD':       False,
    }
    try:
        with open('save.txt', 'r') as file:
            pass
    except:
        print('No saved file, using defaults')
        return DEF
    with open('save.txt', 'r') as file:
        li = file.read().split(';')
        lin = [el.split(':') for el in li]
        DEF = {i[0]:eval(i[1]) for i in lin[:-1]}
        DEF['INTERRUPT'] = False
        DEF['COLORLIST'] = [colHex1, colHex2, colHex3, colHex4, colHex5, colHex6]
        return DEF

if __name__ == '__main__':

    app = QApplication([])
    DEFAULTS = initVars()
    w = MainWindow(**DEFAULTS)
    app.exec()



from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import numpy as np
import time


##==============Workers============##
##=================================##
# These fellas do little tasks ON A SINGLE SHARED ARRAY
# The array updaters all inherit the handler, so they can directly maniupalate the array
class Handler(QObject):
    arraySig = pyqtSignal(np.ndarray, int)
    arrayinitSig = pyqtSignal(np.ndarray)
    arrayfpsSig = pyqtSignal(float)
    startSig = pyqtSignal()
    nextSig = pyqtSignal()
    error = pyqtSignal(str)
    ARRAY = []      # Array, shared among workers
    ARRAYOLD = []      # Array, shared among workers
    BOUNDARY = []

    def __init__(self, **kwargs):
        """ Controls workers for the array updates,
            and processes the arrays returned. """
        QObject.__init__(self)
        self.resize_array(kwargs['THRESHOLD'], kwargs['N'], kwargs['D'])
        self.fpsRoll = np.zeros(5, float)

#   def updater_start(self, frame1, frame2):
#       self.process(frame1)
#       self.arrayinitSig.emit(Handler.ARRAY)
#       self.process(frame2)

    def next_array(self, frame):
        now = time.time()
        self.arraySig.emit(Handler.ARRAY, frame['wolfpos'])
        self.process(frame)
        self.fpsRoll[0] = time.time()-now
        self.fpsRoll = np.roll(self.fpsRoll, 1)
        self.arrayfpsSig.emit(np.mean(self.fpsRoll))

    def process(self, job):
        if job['wolfpole'] >= 0:
            self.clear_wavefront(job['wolfpos'], job['wolfscale'], job['wolfpole'])
        for i in range(job['noisesteps']):
            self.resize_array(job['threshold'], job['n'], job['d'])
        self.set_boundary(job['ub'], job['rb'], job['db'], job['lb'])
        if job['isingupdates'] > 0:
            self.ising_process(job['isingupdates'], job['beta'])
        if not job['conwayrules'] == []:
            self.conway_process(job['conwayrules'])

    def set_boundary(self, ub, rb, db, lb):
        if ub >= 0:
            Handler.ARRAY[0, ...] = ub
        if db >= 0:
            Handler.ARRAY[-1, ...] = db
        if lb >= 0:
            Handler.ARRAY[..., 0] = lb
        if rb >= 0:
            Handler.ARRAY[..., -1] = rb

    def clear_wavefront(self, start, scale, polarity):
        n = Handler.ARRAY.shape[0]
        for i in range(scale):
            Handler.ARRAY[(start + i) % n, ...] = polarity

    def resize_array(self, threshold, height, width):
        Handler.ARRAY = np.zeros([height, width], bool)
        Handler.ARRAYOLD = np.zeros([height, width], bool)
        self.noise_process(threshold)

    def noise_process(self, threshold):
        A = np.random.random(Handler.ARRAY.shape) > threshold
        B = np.bitwise_xor(Handler.ARRAY, A)
        Handler.ARRAY = B

    def ising_process(self, updates, beta):
        cost = np.zeros(3, float)
        cost[1] = np.exp(-4 * beta)
        cost[2] = cost[1] ** 2
        # Really want to undestand object permanence better, here is a case-study--
        # this breaks without np.copy. Maybe werite all the array engines in Cython so
        # its all a little more clear.
       #A = Handler.ARRAY
        A = np.copy(Handler.ARRAY)
        N = A.shape[0]
        D = A.shape[1]
        for _ in range(updates):
            a = np.random.randint(N)
            b = np.random.randint(D)
            nb = np.sum([A[a][b] == A[(a + 1) % N][b],
                            A[a][b] == A[(a - 1) % N][b],
                            A[a][b] == A[a][(b + 1) % D],
                            A[a][b] == A[a][(b - 1) % D],
                            -2])
            if nb <= 0 or np.random.random() < cost[nb]:
                A[a][b] = not A[a][b]
        Handler.ARRAY = A

    def conway_process(self, rule):
        A = Handler.ARRAY
        l = np.roll(A, -1, axis=0)
        r = np.roll(A, 1, axis=0)
        u = np.roll(A, 1 , axis=1)
        d = np.roll(A, -1, axis=1)
        ul = np.roll(l, 1, axis=1)
        dl = np.roll(l, -1, axis=1)
        ur = np.roll(r, 1, axis=1)
        dr = np.roll(r, -1, axis=1)
        NB = np.zeros(A.shape) + l + r + u + d + ul + dl + ur + dr
        #cells still alive after rule 1
        rule1 = np.bitwise_and(A, NB >= rule[0])
        #alive cells that will live
        rule2 = np.bitwise_and(rule1, NB <= rule[1])
        #dead cells that rebirth
        rule4 = np.bitwise_and(~A, NB >= rule[2])
        rule5 = np.bitwise_and(rule4, NB <= rule[3])
        #should just be the live cells
        Handler.ARRAY = rule2 + rule5
        # Rule 4 was being added in here for ages! Produces mad shapes!
        # This  is equivalent to makeing #4=8



from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import numpy as np
import time


##===============TaskManager===============##
##=========================================##
# This fancy chap goes into the thread with all the workers and controls what they do
# Also inherits the array handler so it can send and recieve arrays.
# All signals come in and out of this guy.
# TODO update the taskman
class RunController(QObject):
    frameSig = pyqtSignal(int)
    finished = pyqtSignal()
    handlerSig = pyqtSignal(dict)
    handlerinitSig = pyqtSignal(dict, dict)
#   isingSig = pyqtSignal(int, float)
#   noiseSig = pyqtSignal(float)
#   conwaySig = pyqtSignal(list)
#   clearSig = pyqtSignal(float, int, int)
#   boundSig = pyqtSignal(int)
#   waveSig = pyqtSignal(int, int, int)
    breakSig = pyqtSignal()
    error = pyqtSignal(str)

    def __init__(self, **kwargs):
        """Run controller makes sure the run doesnt get out of hand"""
        QObject.__init__(self)
        self.st = kwargs
        self.fpsTimer = QTimer(self)
        # Maximum fps = 1000 / the following
        self.fpsTimer.setInterval(1)
        self.mainTime = QTimer(self)
        self.wavecounter = 0
        # How often does a dynamic run break to look for new settings?
        self.mainTime.setInterval(99000)

#==============Changes the internal settings================#
# Should make the event queue feeding this baby LIFO
    def change_settings(self, kwargs):
        for i in kwargs:
            self.st[i] = kwargs[i]

#===============MAIN PROCESS OF THE THREAD===================#
    def process(self):
        self.error.emit('Process Starting!')
        self.mainTime.start()
#       frame1 = self.prepare_frame()
#       frame2 = self.prepare_frame()
#       self.handlerinitSig.emit(frame1, frame2)
        self.frame = self.prepare_frame()
        self.frametime = time.time()
        self.next_frame()

    def next_frame(self):
        QCoreApplication.processEvents()
        while time.time() - self.frametime < self.st['SPEED']:
            QThread.msleep(1)
        if 0 <= self.st['IMAGEUPDATES'] <= self.st['RUNFRAMES']:
            self.error.emit('Run finished')
            self.finished.emit()
            return
        if QThread.currentThread().isInterruptionRequested():
            self.error.emit('Interrupted')
            self.finished.emit()
            return
        if not self.st['RUN'] and not self.st['EQUILIBRATE'] and not self.st['CLEAR']:
            self.error.emit('Nothing happening!')
            self.finished.emit()
            return
        self.handlerSig.emit(self.frame)
        self.frame = self.prepare_frame()
        self.frametime = time.time()

    def prepare_frame(self):
        frame={
            'n':self.st['N'],
            'd':self.st['D'],
            'threshold':self.st['THRESHOLD'],
            'noisesteps':0,
            'isingupdates':0,
            'conwayrules':[],
            'beta':self.st['BETA'],
            'ub':self.st['UB'],
            'rb':self.st['RB'],
            'db':self.st['DB'],
            'lb':self.st['LB'],
            'wolfpole':self.st['WOLFPOLARITY'],
            'wolfpos':0,
            'wolfscale':self.st['WOLFSCALE'],
            }
        self.wavecounter += self.st['WOLFSCALE']
        self.wavecounter %= self.st['N']
        frame['wolfpos'] = self.wavecounter
        if self.st['WOLFWAVE'] == 0:
            frame['wolfpole'] = -1
        if self.st['EQUILIBRATE']:
            frame['conwayrules'] = []
            frame['isingupdates'] = self.st['LONGNUM']
        if self.st['CLEAR']:
            frame['conwayrules'] = []
            frame['isingupdates'] = 0
            frame['noisesteps'] = 1
            self.st['RUNFRAMES'] += 1
        if self.st['RUN']:
            rules = self.st['RULES']
            if len(rules) == 0:
                rule = []
            else:
                rule = rules[self.st['RUNFRAMES'] % len(rules)]
            if self.st['CONWAY'] and self.st['STOCHASTIC']:
                if self.st['RUNFRAMES'] % 2:
                    frame['isingupdates'] = self.st['MONTEUPDATES']
                else:
                    frame['conwayrules'] = rule
            else:
                frame['isingupdates'] = self.st['MONTEUPDATES'] * self.st['STOCHASTIC']
                frame['conwayrules'] = rule
            self.st['RUNFRAMES'] += 1
            self.frameSig.emit(self.st['RUNFRAMES'])
        return frame



from PyQt5.QtGui import *

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from functools import partial

import numpy as np
import time

class ImageCreator(QObject):
    imageSig = pyqtSignal(QPixmap)
    nextarraySig = pyqtSignal()
    breakSig = pyqtSignal()
    error = pyqtSignal(str)
    finished = pyqtSignal()
    canvasfpsSig = pyqtSignal(float)

    def __init__(self, **kwargs):
        QObject.__init__(self)

        self.error.emit('Image thread starting up!')
        self.colorList = kwargs['COLORLIST']
        self.degree = 2

        self.scale = kwargs['SCALE']
        self.fpsRoll = np.zeros(5, float)
        self.kwargs = kwargs

        self.N = kwargs['N']
        self.D = kwargs['D']
        self.resize_array(self.N, self.D)

        line = np.random.randint(0, 2, (self.N))
        self.wolf = self.wolframgen(line)
        self.wavecounter = 0
        self.savecount = 0

    def reset_gifcount(self):
        self.savecount = 0

    # Resize/reset
    def resize_array(self, N, D):
        self.ARRAY = np.zeros([N, D], bool)
        self.ARRAYOLD = np.zeros([N, D], bool)
        self.LIVING = np.zeros([0, 2], bool)
        self.CHANGE = np.zeros([0, 3], bool)
        self.image = QImage(N, D, QImage.Format_ARGB32)
        self.export_array(self.ARRAY)

#==============Changes the internal settings================#
# Should make the event queue feeding this baby LIFO
    def change_settings(self, kwargs):
        line = np.random.randint(0, 2, (self.N))
        self.wolf = self.wolframgen(line)
        for i in kwargs:
            self.kwargs[i] = kwargs[i]
        self.addColors()

    def addColors(self):
        self.colorList = self.kwargs['COLORLIST']
        self.degree = self.kwargs['DEGREE']

#==============Wolfram-style Cellular Automata==============#
    def wolfram_scroll(self):
        n = int(self.D / self.kwargs['WOLFSCALE'])
        line = next(self.wolf)
        [self.image.setPixel((self.wavecounter + j) % self.N, i,\
            self.colorList[line[int(i / self.kwargs['WOLFSCALE']) % n] + 2])
                for i in range(self.D) for j in range(self.kwargs['WOLFSCALE'])]

    def wolframgen(self, line):
        n = int(self.D / self.kwargs['WOLFSCALE'])
      # n = self.shape
        rule = str(bin(self.kwargs['WOLFRULE']))[2:]
        while len(rule) < 8:
            rule = '0' + rule
        while True:
            nb = [int(str(line[(i - 1) % n]) + str(line[i]) + str(line[(i + 1) % n]),\
                    2) for i in range(n)]
            line = [int(rule[-i - 1]) for i in nb]
            yield line

    def wolfram_paint(self):
        D = int(self.D / self.kwargs['WOLFSCALE'])
        N = int(self.N / self.kwargs['WOLFSCALE'])
        im = QImage(N, D, QImage.Format_ARGB32)
      # line = np.random.randint(0, 2, (shape))
        line = np.zeros(D, int)
        line[int(D / 2)] = 1
        linegen = self.wolframgen(line)
        for idx, lin in enumerate(range(N)):
            line = next(linegen)
            for idy, pix in enumerate(line):
                im.setPixel(idx % N, idy, self.colorList[pix + 2])      #pix+2 means background colors
            if idx == N:
                break
        self.send_image(im)

#===============Array processing and Image export=============#
    def send_image(self, image):
        ims = image.scaled(QSize(self.N * self.scale, self.D * self.scale))
        nupix = QPixmap()
        nupix.convertFromImage(ims)
        self.imageSig.emit(nupix)
        if self.kwargs['RECORD']:
            ims.save('images/temp{:>04d}.png'.format(self.savecount), 'PNG')
            self.savecount += 1

#   def processer_start(self, array):
#       self.process_array(array)
#       self.nextarraySig.emit()

    def process(self, array, pos):
        self.wavecounter = pos
        self.send_image(self.image)
        self.process_array(array)
        self.nextarraySig.emit()

    def process_array(self, array):
        now = time.time()
        self.ARRAY = array
        self.update_change()
        self.export_list(self.CHANGE)
        if self.kwargs['WOLFWAVE']:
            self.wolfram_scroll()
        self.ARRAYOLD = np.copy(self.ARRAY)
        self.fpsRoll[0] = time.time()-now
        self.fpsRoll = np.roll(self.fpsRoll, 1)
        self.canvasfpsSig.emit(np.mean(self.fpsRoll))

    def update_living(self):
        self.LIVING = np.argwhere(self.ARRAY)

    def update_change(self):
        common = np.bitwise_and(self.ARRAY, self.ARRAYOLD)
        onlyOld = np.bitwise_xor(common, self.ARRAYOLD)
        onlyNew = np.bitwise_xor(common, self.ARRAY)
        births = np.argwhere(onlyNew)
        deaths = np.argwhere(onlyOld)
        b = np.concatenate((births, np.ones([births.shape[0], 1], int)), axis=1)
        d = np.concatenate((deaths, np.zeros([deaths.shape[0], 1], int)), axis=1)
        self.CHANGE = np.concatenate((b, d))
#       if not self.CHANGE.size:
#           self.breakSig.emit()

    # Updates image with values from entire array. SLOW
    def export_array(self, A):
        im = self.image
        for i in range(A.shape[0]):
            for j in range(A.shape[1]):
                num = int(A[i][j])
                color = self.colorList[num]
                im.setPixel(i, j, color)

    # Updates image only where the pixels have changed. FASTER
    def export_list(self, L):
        [self.image.setPixel(el[0], el[1], self.colorList[el[2]]) for el in L]

#====================The canvas=================#
# It isnt so exciting since I moved everything to the processor
class Canvas(QLabel):
    canvasfpsSig = pyqtSignal(float)

    def initialize(self, **kwargs):
        self.primaryColor = QColor(kwargs['COLORLIST'][0])
        self.colorList = []
        self.degree = 2
        self.N = kwargs['N']
        self.D = kwargs['D']
        self.scale = kwargs['SCALE']
        self.reset()

    def reset(self):
        self.setPixmap(QPixmap(self.N * self.scale, self.D * self.scale))
        self.pixmap().fill(self.primaryColor)

    def paint(self, image):
        self.setPixmap(image)
        self.repaint()



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='LatticeModelMashup',
    version='0.98.2',
    packages=['latticeModelMashup',],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/fergusfettes/lattice',
    author='Fergus Fettes',
    author_email='fergusfettes@gmail.com',
    description='An ungodly mixture of lattice models',
    install_requires=[
        'imageio',
        'numpy',
        'python-qt5',
        'ffmpeg-python',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Topic :: Games/Entertainment :: Simulation',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX :: Linux',
    ],
    entry_points={
        'console_scripts': [
            'latticeMash = latticeModelMashup.__init__:main',
        ],
    },
)



from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from functools import partial

from engineOperator import *
from imageProcessing import *

import random as ra
import math
import ffmpeg
import os
import re

# Draws the main window and contains the simulation code
class MainWindow(QWidget):

    # Initialises the window and variables. Very uncertain about which
    # variables should go where, but it's fine for now I guess.
    def __init__(self, **DEFAULTS):
        super().__init__()

        # Internal Vars
        # This is the sigmoid for converting the 'coverage percent' (lie) to a threshold.
        # To make it less steep (ie less flat at the top and bottom), reduce the #24.
        self.threshval = list(map((lambda x:1 - (1 / (1 + math.exp(-(x - 0.5) * 24)))), np.arange(100) / 100))
        self.conwayMangled = False
        # Degree of the Potts model
###     self.deg = DEFAULTS['DEGREE']   # Will the poor old Potts model ever make a comeback? Let's see.

        # Save kwargs
        self.kwargs = DEFAULTS

        # INITS
        self.kwarg_send_timer = QTimer()
        self.kwarg_send_timer.timeout.connect(self.kwarg_send)
        # This means that you have to stop manipulating the controls for a full three
        # seconds for the changes to be sent to the thread. This stops uneccesary
        # restarts, but needs to be tuned for comfort.
        self.kwarg_send_timer.setInterval(500)
        self.initGUI(**DEFAULTS)

#=====================Settings controllers================#
    def Kwarger(self, kwarg, callback):
        self.changeKwarg(kwarg, callback())

    def changeKwarg(self, kwarg, nuVal):
        print('Changing {}'.format(kwarg))
        self.kwargs[kwarg] = nuVal
        print(self.kwargs[kwarg])
        self.kwarg_send_timer.start()

    def kwarg_send(self):
        print('Sending new kwargs')
        self.engine.update_kwargs(**self.kwargs)
        self.kwarg_send_timer.stop()

    def frameChange(self):
        self.changeKwarg('IMAGEUPDATES', self.frameCtrl.value())

    def speedChange(self):
        self.changeKwarg('SPEED', 1 / self.speedCtrl.value())
        self.speedLabel.setText('Max FPS = {:d}'.format(int(self.speedCtrl.value())))

    def coverageChange(self):
        coverage = self.thresholdCtrl.value()
        self.changeKwarg('THRESHOLD', self.threshval[coverage])
        self.thresholdLabel.setText('Threshold = {:2.2f}'.format(self.threshval[coverage]))

    def sliderChange(self):
        self.changeKwarg('BETA', self.tempCtrl.value() / 100)
        self.tempLabel.setText('Beta = {:01.2f}'.format(self.tempCtrl.value() / 100))

    def choose_color(self, callback, *args):
        dlg = QColorDialog()
        if dlg.exec():
            callback(dlg.selectedColor().name(), *args)

    def set_color(self, hexx, button, num):
        templist = self.kwargs['COLORLIST']
        templist[num] = QColor(hexx).rgba()
        self.changeKwarg('COLORLIST', templist)
        button.setStyleSheet('QPushButton { background-color: %s; }' % hexx)

    def conway_mangler(self):
        if self.conwayMangled:
            old = self.conwayRules.toPlainText()
            self.conwayRules.setText(old[2:-2])
            self.conwayMangled = False
        else:
            old = self.conwayRules.toPlainText()
            self.conwayRules.setText('!!{}!!'.format(old))
            self.conwayMangled = True

    def record_change(self):
        value = not self.kwargs['RECORD']
        self.changeKwarg('RECORD', value)
        if value:
            self.record.setStyleSheet('QPushButton { background-color: %s; }' %  \
                                            QColor(Qt.red).name())
        else:
            self.record.setStyleSheet('QPushButton { background-color: %s; }' %  \
                                            QColor(Qt.white).name())
            self.gif_creator()
            self.engine.reset_gifcount()


    def rulesChange(self):
        regexTestString=r'^(?:([0-9])(?:,\ ?)([0-9])(?:,\ ?)([0-9])(?:,\ ?)([0-9])(?:;\ ?)[\ \n]*)+$'
        regexMatchString=r'([0-9])(?:,\ ?)([0-9])(?:,\ ?)([0-9])(?:,\ ?)([0-9])(?:;\ ?)'
        text = self.conwayRules.toPlainText()
        strTest = re.match(regexTestString, text)
        # This timer waits for ten seconds for you to dick about with the rules before
        # sending them to the engine. If you get caught with your pants down, conway will
        # turn off for some seconds.
        if strTest is None:
            self.conwayPalette.setColor(QPalette.Base, Qt.red)
            self.conwayRules.setPalette(self.conwayPalette)
            ruleIter = re.finditer(regexTestString, text)
            self.rul = [i.group(1,2,3,4) for i in ruleIter]
        else:
            self.conwayPalette.setColor(QPalette.Base, Qt.green)
            self.conwayRules.setPalette(self.conwayPalette)
            ruleIter = re.finditer(regexMatchString, text)
            self.rul = [i.group(1,2,3,4) for i in ruleIter]
        self.send_rule()

    def send_rule(self):
        print('Rule sending!')
        rules = [[int(j) for j in i] for i in self.rul]
        self.changeKwarg('RULES', rules)
        self.changeKwarg('CONWAY', not rules == [])

#=====================Save defaults and GUI ket controls===============#
    def gif_creator(self):
        filenums = [re.findall('([0-9]{3}).png', i) for i in os.listdir('images')]
        fileints = [int(i[0]) for i in filter(None, filenums)]
        rules = ''.join([''.join([str(i) for i in j]) for j in self.kwargs['RULES']])

# Watermark, off for now
#       overlay_file = ffmpeg.input('images/watermark.png')
#       (
#           ffmpeg
#           .input('images/temp{:04d}.png'.format(max(fileints)))
#           .overlay(overlay_file)
#           .output('images/temp{:04d}.png'.format(max(fileints)))
#           .overwrite_output()
#           .run()
#       )
        (
            ffmpeg
            .input('images/temp%04d.png')
            .output('images/{3}x{4}-frames:{2}-wolf:{1}-rule:{0}.gif'.format(
                rules, self.kwargs['WOLFWAVE'], max(fileints),
                self.kwargs['D'], self.kwargs['N']), framerate=5, f='gif')
            .run()
        )

        for i in range(max(fileints) + 1):
            os.remove('images/temp{:04d}.png'.format(i))


    #TODO: make it save the previous configuration before overwriting
    def save_defaults(self):
        with open('save.txt', 'w') as file:
            savestr = ''.join(['{0}:{1};'.format(i, self.kwargs[i]) for i in self.kwargs])
            file.write(savestr)

    def keyPressEvent(self, e):
        # TODO: make this a dictonary
        if e.key() == Qt.Key_Escape:
            if self.engine.updatethread.isRunning() or self.engine.imagethread.isRunning()\
                    or self.engine.taskthread.isRunning():
                self.changeKwarg('EQUILIBRATE', False)
                self.changeKwarg('CLEAR', False)
                self.changeKwarg('RUN', False)
                self.engine.taskthread.requestInterruption()
                self.engine.updatethread.requestInterruption()
                self.engine.imagethread.quit()
                print('Attempting to interrupt!')
                # TODO: add a 'interrupted by user' popup (after a 'interrupting!'?)
            else:
                self.engine.taskthread.deleteLater()
                self.engine.updatethread.deleteLater()
                self.engine.imagethread.deleteLater()
                QCoreApplication.instance().quit()
                print('Threads shutting down and powering off')
                # TODO: add a 'are you sure?' popup
        # left alt key. guess i could just look this up?
        elif e.key() == 16777251:
            self.speedCtrl.setFocus()
        elif e.key() == Qt.Key_C:
            self.speedCtrl.triggerAction(QSlider.SliderPageStepAdd)
        elif e.key() == Qt.Key_X:
            self.speedCtrl.triggerAction(QSlider.SliderPageStepSub)
        elif e.key() == Qt.Key_B:
            self.background.click()
        elif e.key() == Qt.Key_L:
            self.equilibrate.click()
        elif e.key() == Qt.Key_Z:
            self.short.click()
        elif e.key() == Qt.Key_D:
            self.tempCtrl.triggerAction(QSlider.SliderPageStepAdd)
        elif e.key() == Qt.Key_A:
            self.tempCtrl.triggerAction(QSlider.SliderPageStepSub)
        elif e.key() == Qt.Key_W:
            self.thresholdCtrl.triggerAction(QSlider.SliderPageStepAdd)
        elif e.key() == Qt.Key_S:
            self.thresholdCtrl.triggerAction(QSlider.SliderPageStepSub)
        # Change Colors, RF
        elif e.key() == Qt.Key_R:
            self.primaryButton.click()
        elif e.key() == Qt.Key_F:
            self.secondaryButton.click()
        elif e.key() == Qt.Key_1:
            state = self.stochasticBox.isChecked()
            self.stochasticBox.setChecked(not state)
        elif e.key() == Qt.Key_2:
            self.conway_mangler()
        elif e.key() == Qt.Key_E:
            self.dynamic.click()
        elif e.key() == Qt.Key_Q:
            self.clear.click()

##==========================THEGUI==============================##
##==============================================================##
    # Initialise GUI
    def initGUI(self, **DEFAULTS):
        # Initialise the thread manager and painter and feed them the UI elements they
        # need to control. TODO: cleaner way of connecting signals to a parent? Using
        # Super perhaps? TODO: pass the guys with args anyway no?
        self.canvas = Canvas()
        self.canvas.initialize(**DEFAULTS)
        self.frameLabel = QLabel()
        self.arrayfpsLabel = QLabel()
        self.canvasfpsLabel = QLabel()
        self.engine = EngineOperator(self.canvas, self.frameLabel, self.arrayfpsLabel,
                                     self.canvasfpsLabel, **DEFAULTS)

##====================Controlls on the left====================##
        # Main pushbutton, this should be better highlighted.
        self.dynamic = QPushButton()
        self.dynamic.setText('Dynamic')
        self.dynamic.setStyleSheet('QPushButton { background-color: %s; }' %  \
                                         QColor('#ffff00ff').name())
        self.dynamic.clicked.connect(self.engine.dynamic_run)

        # Buttons and slier in the top left
        self.short = QPushButton()
        self.equilibrate = QPushButton()
        self.short.setText('Step')
        self.short.clicked.connect(self.engine.static_run)
        self.equilibrate.setText('Long')
        self.equilibrate.clicked.connect(self.engine.long_run)
        self.clear = QPushButton()
        self.clear.setText('Clear')
        self.clear.clicked.connect(self.engine.clear_array)
        self.background = QPushButton()
        self.background.setText('Background')
        self.background.clicked.connect(self.engine.clear_background)
        self.thresholdCtrl = QSlider(Qt.Vertical)
        self.thresholdCtrl.setTickPosition(QSlider.TicksRight)
        self.thresholdCtrl.setTickInterval(20)
        self.thresholdCtrl.setMinimum(0)
        self.thresholdCtrl.setMaximum(99)
        self.thresholdCtrl.setPageStep(4)
        self.thresholdCtrl.setValue(26)
        self.thresholdCtrl.valueChanged.connect(self.coverageChange)
        self.thresholdLabel= QLabel()
        self.thresholdLabel.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.thresholdLabel.setText('Threshold = {:2.2f}'.format(self.kwargs['THRESHOLD']))
        # 'toplefthbox'
        tlhb = QHBoxLayout()
        tlhbvb = QVBoxLayout()
        tlhbvb.addWidget(self.short)
        tlhbvb.addWidget(self.equilibrate)
        tlhbvb.addWidget(self.clear)
        tlhbvb.addWidget(self.background)
        tlhbvb.addWidget(self.thresholdLabel)
        tlhb.addLayout(tlhbvb)
        tlhb.addWidget(self.thresholdCtrl)

        # Conway rule selector
        self.conwayLabel = QLabel()
        self.conwayRules = QTextEdit()
        self.conwayRules.setMaximumSize(200, 100)
        self.textHolder = QVBoxLayout()
        self.textHolder.addWidget(self.conwayLabel)
        self.textHolder.addWidget(self.conwayRules)
        self.conwayLabel.setText('Conway Rules')
        self.conwayRules.setText(''.join(['{0},{1},{2},{3};'.format(*i) for i in self.kwargs['RULES']]))
        self.conwayPalette = self.conwayRules.palette()
        self.conwayPalette.setColor(QPalette.Base, Qt.green)
        self.conwayRules.setPalette(self.conwayPalette)
        self.conwayRules.setAcceptRichText(False)
        self.conwayRules.textChanged.connect(self.rulesChange)

        # Rules controller for cellular automata
        self.automataTop = QLabel('Automata Rules')
        self.automataLabel = QLabel('Rule')
        self.wolfRule = QSpinBox()
        self.wolfRule.setRange(0, 255)
        self.wolfRule.setSingleStep(1)
        self.wolfRule.setValue(self.kwargs['WOLFRULE'])
        self.wolfRule.setMaximumSize(100, 40)
        self.wolfRule.valueChanged.connect(partial(self.Kwarger, 'WOLFRULE', self.wolfRule.value))
        self.wolfScaleLabel = QLabel('Scale')
        self.wolfScale = QSpinBox()
        self.wolfScale.setRange(1, 100)
        self.wolfScale.setSingleStep(1)
        self.wolfScale.setValue(self.kwargs['WOLFSCALE'])
        self.wolfScale.setMaximumSize(100, 40)
        self.wolfScale.valueChanged.connect(partial(self.Kwarger, 'WOLFSCALE', self.wolfScale.value))
        self.wolfCheck = QCheckBox('WolfWaveTM')
        self.wolfCheck.setChecked(self.kwargs['WOLFWAVE'])
        self.wolfCheck.stateChanged.connect(partial(self.Kwarger, 'WOLFWAVE', self.wolfCheck.isChecked))
        vbWolf = QVBoxLayout()
        hbt = QHBoxLayout()
        hbt.addWidget(self.automataLabel)
        hbt.addWidget(self.wolfRule)
        hbt1 = QHBoxLayout()
        hbt1.addWidget(self.wolfScaleLabel)
        hbt1.addWidget(self.wolfScale)
        vbWolf.addWidget(self.automataTop)
        vbWolf.addLayout(hbt)
        vbWolf.addLayout(hbt1)
        vbWolf.addWidget(self.wolfCheck)
        rulehb = QHBoxLayout()
        rulehb.addLayout(self.textHolder)
        vlineR = QFrame()
        vlineR.setFrameShape(QFrame.VLine)
        vlineR.setFrameShadow(QFrame.Sunken)
        rulehb.addWidget(vlineR)
        rulehb.addLayout(vbWolf)

        # Boundary conditions
        self.UBL = QLabel('UB')
        self.UB = QSpinBox()
        self.UB.setRange(-1, 1)
        self.UB.setValue(self.kwargs['UB'])
        self.UB.setMaximumSize(40, 40)
        self.UB.valueChanged.connect(partial(self.Kwarger, 'UB', self.UB.value))
        hlineUB = QFrame()
        hlineUB.setFrameShape(QFrame.HLine)
        hlineUB.setFrameShadow(QFrame.Sunken)
        self.LBL = QLabel('LB')
        self.LB = QSpinBox()
        self.LB.setRange(-1, 1)
        self.LB.setValue(self.kwargs['LB'])
        self.LB.setMaximumSize(40, 40)
        self.LB.valueChanged.connect(partial(self.Kwarger, 'LB', self.LB.value))
        vlineLB = QFrame()
        vlineLB.setFrameShape(QFrame.VLine)
        vlineLB.setFrameShadow(QFrame.Sunken)
#       self.boundPole.setMaximumSize(100, 40)
        self.wolfLabel = QLabel('WB')
        self.wolfPole = QSpinBox()
        self.wolfPole.setRange(-1, 1)
        self.wolfPole.setValue(self.kwargs['WOLFPOLARITY'])
        self.wolfPole.setMaximumSize(40, 40)
        self.wolfPole.valueChanged.connect(partial(self.Kwarger, 'WOLFPOLARITY', self.wolfPole.value))
        vlineWB = QFrame()
        vlineWB.setFrameShape(QFrame.VLine)
        vlineWB.setFrameShadow(QFrame.Sunken)
        self.RBL = QLabel('RB')
        self.RB = QSpinBox()
        self.RB.setRange(-1, 1)
        self.RB.setMaximumSize(40, 40)
        self.RB.setValue(self.kwargs['RB'])
        self.RB.valueChanged.connect(partial(self.Kwarger, 'RB', self.RB.value))
        vlineRB = QFrame()
        vlineRB.setFrameShape(QFrame.VLine)
        vlineRB.setFrameShadow(QFrame.Sunken)
        self.DBL = QLabel('DB')
        self.DB = QSpinBox()
        self.DB.setRange(-1, 1)
        self.DB.setValue(self.kwargs['DB'])
        self.DB.setMaximumSize(40, 40)
        self.DB.valueChanged.connect(partial(self.Kwarger, 'DB', self.DB.value))
        hlineDB = QFrame()
        hlineDB.setFrameShape(QFrame.HLine)
        hlineDB.setFrameShadow(QFrame.Sunken)
        grPole = QGridLayout()
        grPole.addWidget(self.UBL, 0, 2)
        grPole.addWidget(self.UB, 0, 3)
        grPole.addWidget(hlineUB, 1, 1, 1, 5)
        grPole.addWidget(self.LBL, 2, 0)
        grPole.addWidget(self.LB, 3, 0)
        grPole.addWidget(vlineLB, 2, 1, 3, 1)
        grPole.addWidget(self.wolfLabel, 2, 2)
        grPole.addWidget(self.wolfPole, 3, 2)
        grPole.addWidget(vlineWB, 2, 3, 3, 1)
        grPole.addWidget(self.RBL, 2, 6)
        grPole.addWidget(self.RB, 3, 6)
        grPole.addWidget(vlineRB, 2, 5, 3, 1)
        grPole.addWidget(self.DBL, 6, 2)
        grPole.addWidget(self.DB, 6, 3)
        grPole.addWidget(hlineDB, 5, 1, 1, 5)



        # Default value changer
        NLab = QLabel('NxD: ')
        NLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.NCtrl = QSpinBox()
        self.NCtrl.setRange(10, 1000)
        self.NCtrl.setSingleStep(10)
        self.NCtrl.setValue(self.kwargs['D'])
        self.NCtrl.setMaximumSize(100, 40)
        self.NCtrl.valueChanged.connect(partial(self.Kwarger, 'D', self.NCtrl.value))
        self.DCtrl = QSpinBox()
        self.DCtrl.setRange(10, 1000)
        self.DCtrl.setSingleStep(10)
        self.DCtrl.setValue(self.kwargs['N'])
        self.DCtrl.setMaximumSize(100, 40)
        self.DCtrl.valueChanged.connect(partial(self.Kwarger, 'N', self.DCtrl.value))
        MonteUpLab = QLabel('Monte= ')
        MonteUpLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.MonteUpCtrl = QSpinBox()
        self.MonteUpCtrl.setRange(100, 5000)
        self.MonteUpCtrl.setSingleStep(100)
        self.MonteUpCtrl.setValue(self.kwargs['MONTEUPDATES'])
        self.MonteUpCtrl.setMaximumSize(100, 40)
        self.MonteUpCtrl.valueChanged.connect(partial(self.Kwarger, 'MONTEUPDATES',
                                                      self.MonteUpCtrl.value))
        LongLab = QLabel('Long#= ')
        LongLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.LongCtrl = QSpinBox()
        self.LongCtrl.setRange(10000, 1000000)
        self.LongCtrl.setSingleStep(10000)
        self.LongCtrl.setValue(self.kwargs['EQUILIBRATE'])
        self.LongCtrl.setMaximumSize(100, 40)
        self.LongCtrl.valueChanged.connect(partial(self.Kwarger, 'EQUILIBRATE',
                                                   self.LongCtrl.value))
        DegreeLab = QLabel('Degree= ')
        DegreeLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.DegreeCtrl = QSpinBox()
        self.DegreeCtrl.setRange(2, 10)
        self.DegreeCtrl.setSingleStep(1)
        self.DegreeCtrl.setValue(self.kwargs['DEGREE'])
        self.DegreeCtrl.setMaximumSize(100, 40)
        self.DegreeCtrl.valueChanged.connect(partial(self.Kwarger, 'DEGREE',
                                                     self.DegreeCtrl.value))
        ScaleLab = QLabel('Scale= ')
        ScaleLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.ScaleCtrl = QSpinBox()
        self.ScaleCtrl.setRange(1, 20)
        self.ScaleCtrl.setSingleStep(1)
        self.ScaleCtrl.setValue(self.kwargs['SCALE'])
        self.ScaleCtrl.setMaximumSize(100, 40)
        self.ScaleCtrl.valueChanged.connect(partial(self.Kwarger, 'SCALE',
                                                    self.ScaleCtrl.value))
        SaveDefaults = QPushButton('Save Defaults')
        SaveDefaults.clicked.connect(self.save_defaults)
        defGr = QGridLayout()
        defGr.addWidget(NLab, 0, 0)
        defGr.addWidget(self.NCtrl, 0, 1)
        defGr.addWidget(self.DCtrl, 0, 2)
        defGr.addWidget(MonteUpLab, 1, 0)
        defGr.addWidget(self.MonteUpCtrl, 1, 1, 1, 2)
        defGr.addWidget(LongLab, 0, 3)
        defGr.addWidget(self.LongCtrl, 0, 4)
        defGr.addWidget(DegreeLab, 1, 3)
        defGr.addWidget(self.DegreeCtrl, 1, 4)
        defGr.addWidget(ScaleLab, 2, 0)
        defGr.addWidget(self.ScaleCtrl, 2, 1, 1, 2)
        defGr.addWidget(SaveDefaults, 2, 3, 1, 2)

#       self.fixedBox = QCheckBox('Fixed')
#       self.fixedBox.setChecked(True)
#      #self.fixedBox.stateChanged.connect(self.fixedChange)
        # TODO make this work
#       sathb = QHBoxLayout()
#       self.satLab = QLabel('Saturation %')
#       self.satLab.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
#       self.satCtrl = QSpinBox()
#       self.satCtrl.setRange(20, 100)
#       self.satCtrl.setSingleStep(10)
#       self.satCtrl.setValue(100)
#       self.satCtrl.setMaximumSize(100, 40)
#       sathb.addWidget(self.fixedBox)
#       sathb.addWidget(self.satLab)
#       sathb.addWidget(self.satCtrl)

        # Color buttons, two for background, two for update and two for mouse.
        # TODO: rework the whole color organisation throughout.
        self.primaryButton = QPushButton()
        self.primaryButton.setStyleSheet('QPushButton { background-color: %s; }' %
                                         QColor(self.kwargs['COLORLIST'][0]).name())
        self.primaryButton.pressed.connect(
            partial(self.choose_color, self.set_color, self.primaryButton, 0))
        self.secondaryButton = QPushButton()
        self.secondaryButton.setStyleSheet('QPushButton { background-color: %s; }' %
                                           QColor(self.kwargs['COLORLIST'][1]).name())
        self.secondaryButton.pressed.connect(
            partial(self.choose_color, self.set_color, self.secondaryButton, 1))
        self.updateButton = QPushButton()
        self.updateButton.setStyleSheet('QPushButton { background-color: %s; }' %
                                         QColor(self.kwargs['COLORLIST'][2]).name())
        self.updateButton.pressed.connect(
            partial(self.choose_color, self.set_color, self.updateButton, 2))
        self.update2Button = QPushButton()
        self.update2Button.setStyleSheet('QPushButton { background-color: %s; }' %
                                         QColor(self.kwargs['COLORLIST'][3]).name())
        self.update2Button.pressed.connect(
            partial(self.choose_color, self.set_color, self.update2Button, 3))
        self.mouseButton = QPushButton()
        self.mouseButton.setStyleSheet('QPushButton { background-color: %s; }' %
                                         QColor(self.kwargs['COLORLIST'][4]).name())
        self.mouseButton.pressed.connect(
            partial(self.choose_color, self.set_color, self.mouseButton, 4))
        self.mouse2Button = QPushButton()
        self.mouse2Button.setStyleSheet('QPushButton { background-color: %s; }' %
                                         QColor(self.kwargs['COLORLIST'][5]).name())
        self.mouse2Button.pressed.connect(
            partial(self.choose_color, self.set_color, self.mouse2Button, 5))
        self.gr = QGridLayout()
        self.gr.addWidget(self.primaryButton, 0, 0)
        self.gr.addWidget(self.secondaryButton, 1, 0)
        self.gr.addWidget(self.updateButton, 0, 1)
        self.gr.addWidget(self.update2Button, 1, 1)
        self.gr.addWidget(self.mouseButton, 0, 2)
        self.gr.addWidget(self.mouse2Button, 1, 2)

        hline = QFrame()
        hline.setFrameShape(QFrame.HLine)
        hline.setFrameShadow(QFrame.Sunken)
        hline2 = QFrame()
        hline2.setFrameShape(QFrame.HLine)
        hline2.setFrameShadow(QFrame.Sunken)
        hline3 = QFrame()
        hline3.setFrameShape(QFrame.HLine)
        hline3.setFrameShadow(QFrame.Sunken)
        hline4 = QFrame()
        hline4.setFrameShape(QFrame.HLine)
        hline4.setFrameShadow(QFrame.Sunken)
        hline5 = QFrame()
        hline5.setFrameShape(QFrame.HLine)
        hline5.setFrameShadow(QFrame.Sunken)
        hline6 = QFrame()
        hline6.setFrameShape(QFrame.HLine)
        hline6.setFrameShadow(QFrame.Sunken)
        vline = QFrame()
        vline.setFrameShape(QFrame.VLine)
        vline.setFrameShadow(QFrame.Sunken)
        vline2 = QFrame()
        vline2.setFrameShape(QFrame.VLine)
        vline2.setFrameShadow(QFrame.Sunken)

        # Collects all of the above into a column.
        vb = QVBoxLayout()
        vb.addWidget(self.dynamic)
        vb.addLayout(tlhb)
        vb.addWidget(hline)
        vb.addWidget(hline2)
        vb.addLayout(rulehb)
        vb.addLayout(grPole)
        vb.addStretch()
        vb.addWidget(hline3)
        vb.addWidget(hline4)
        vb.addLayout(defGr)
#       vb.addLayout(sathb)
        vb.addStretch()
        vb.addWidget(hline5)
        vb.addWidget(hline6)
        vb.addLayout(self.gr)

##===================Canvas and controls, right======================##
        # Stats at the top of the canvas
        statBox = QHBoxLayout()
        self.canvasfpsLabel.setText('Canvas fps: ')
        self.arrayfpsLabel.setText('Array fps: ')
        self.record = QPushButton()
        self.record.setText('Record')
        self.record.clicked.connect(self.record_change)
        statBox.addWidget(self.canvasfpsLabel)
        statBox.addWidget(self.arrayfpsLabel)
        statBox.addWidget(self.record)

        # Temperature slider and label
        self.tempCtrl = QSlider(Qt.Horizontal)
        self.tempCtrl.setTickPosition(QSlider.TicksAbove)
        self.tempCtrl.setTickInterval(20)
        self.tempLabel = QLabel()
        self.tempCtrl.setMinimum(10)
        self.tempCtrl.setMaximum(200)
        self.tempCtrl.setPageStep(20)
        self.tempCtrl.setValue(self.kwargs['BETA'] * 100)
        self.tempCtrl.valueChanged.connect(self.sliderChange)
        self.tempLabel.setText('Beta = {:01.2f}'.format(self.kwargs['BETA']))

        # Speed slider and label
        # TODO: get this working again
        self.speedCtrl = QSlider(Qt.Horizontal)
        self.speedCtrl.setTickPosition(QSlider.TicksBelow)
        self.speedCtrl.setTickInterval(20)
        self.speedCtrl.setMinimum(1)
        self.speedCtrl.setMaximum(100)
        self.speedCtrl.setValue(int(1 / self.kwargs['SPEED']))
        self.speedCtrl.valueChanged.connect(self.speedChange)
        self.speedLabel = QLabel()
        self.speedLabel.setText('Max FPS = {:03d}'.format(int(1 / self.kwargs['SPEED'])))
        self.frameLabel.setText('0000/ ')
        self.frameCtrl = QSpinBox()
        self.frameCtrl.setRange(-1, 2000)
        self.frameCtrl.setSingleStep(10)
        self.frameCtrl.setMaximumSize(100, 20)
        self.frameCtrl.setValue(self.kwargs['IMAGEUPDATES'])
        self.frameCtrl.valueChanged.connect(self.frameChange)

        # 'right bottom box [LEFT/MID/RIGHT]' -- sorts out the sliders
        rbbL = QVBoxLayout()
        rbbL.addWidget(self.speedLabel)
        rbbL.addWidget(self.tempLabel)
        rbbM = QVBoxLayout()
        rbbM.addWidget(self.speedCtrl)
        rbbM.addWidget(self.tempCtrl)
        rbbR = QVBoxLayout()
        rbbRt = QHBoxLayout()
        rbbRt.addWidget(self.frameLabel)
        rbbRt.addWidget(self.frameCtrl)
        self.stochasticBox = QCheckBox('Stochi')
        self.stochasticBox.setChecked(self.kwargs['STOCHASTIC'])
        self.stochasticBox.stateChanged.connect(partial(self.Kwarger, 'STOCHASTIC',
                                                self.stochasticBox.isChecked))

        rbbR.addLayout(rbbRt)
        rbbR.addWidget(self.stochasticBox)
        rbb = QHBoxLayout()
        rbb.addLayout(rbbL)
        rbb.addLayout(rbbM)
        rbb.addLayout(rbbR)

        # Layout canvas and sliders
        hliner1 = QFrame()
        hliner1.setFrameShape(QFrame.HLine)
        hliner1.setFrameShadow(QFrame.Sunken)
        hliner2 = QFrame()
        hliner2.setFrameShape(QFrame.HLine)
        hliner2.setFrameShadow(QFrame.Sunken)
        hliner3 = QFrame()
        hliner3.setFrameShape(QFrame.HLine)
        hliner3.setFrameShadow(QFrame.Sunken)
        hliner4 = QFrame()
        hliner4.setFrameShape(QFrame.HLine)
        hliner4.setFrameShadow(QFrame.Sunken)
        rb = QVBoxLayout()
        rb.addLayout(statBox)
        rb.addWidget(hliner1)
        rb.addWidget(hliner2)
        rb.addWidget(self.canvas)
        rb.addWidget(hliner3)
        rb.addWidget(hliner4)
        rb.addLayout(rbb)

        # Layout control column and canvasbox
        hb = QHBoxLayout()
        hb.addLayout(vb)
        hb.addWidget(vline)
        hb.addWidget(vline2)
        hb.addLayout(rb)

        self.setLayout(hb)
        self.show()

        # The allows i3 to popup the window (add to i3/config)
        # 'for_window [window_role='popup'] floating enable'
        self.setWindowRole('popup')
